/*****************************************************************************
 *  FTools - Freenet Client Protocol Tools
 *
 *  Copyright (C) 2002 Juergen Buchmueller <juergen@pullmoll.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *****************************************************************************/
#include "ftfcp.h"
#include "ftlog.h"

#define	MAXBUFF	32768

int fcp_new(fcp_t **pf, const char *hostname, int port)
{
	fcp_t *f = calloc(sizeof(fcp_t), 1);

	if (NULL == f) {
		return -1;
	}
	if (0 != sk_outgoing(&f->conn, hostname, port)) {
		free(f);
		*pf = NULL;
		return -1;
	}

	*pf = f;
	return 0;
}

int fcp_free(fcp_t **pf)
{
	fcp_t *f;

	if (NULL == pf) {
		return -1;
	}
	f = *pf;
	if (NULL == f) {
		return 0;
	}

	sk_shut(&f->conn);

	if (NULL != f->uri) {
		free(f->uri);
		f->uri = NULL;
	}
	if (NULL != f->public_key) {
		free(f->public_key);
		f->public_key = NULL;
	}
	if (NULL != f->private_key) {
		free(f->private_key);
		f->private_key = NULL;
	}
	if (NULL != f->meta) {
		free(f->meta);
		f->meta = NULL;
	}
	if (NULL != f->data) {
		free(f->data);
		f->data = NULL;
	}
	free(f);
	*pf = NULL;

	return 0;
}

int fcp_reset(fcp_t *f, int reset_data)
{
	if (NULL != f->uri) {
		free(f->uri);
		f->uri = NULL;
	}

	if (NULL != f->public_key) {
		free(f->public_key);
		f->public_key = NULL;
	}

	if (NULL != f->private_key) {
		free(f->private_key);
		f->private_key = NULL;
	}

	if (0 != reset_data) {
		f->metasize = 0;
		f->metaoffs = 0;
		if (NULL != f->meta) {
			free(f->meta);
			f->meta = NULL;
		}

		f->datasize = 0;
		f->dataoffs = 0;
		if (NULL != f->data) {
			free(f->data);
			f->data = NULL;
		}
	}

	f->seen_node_hello = 0;
	if (NULL != f->seen_protocol) {
		free(f->seen_protocol);
		f->seen_protocol = NULL;
	}
	if (NULL != f->seen_node) {
		free(f->seen_node);
		f->seen_node = NULL;
	}
	f->seen_end_message = 0;
	f->seen_data_found = 0;
	f->seen_uri_error = 0;
	f->seen_restarted = 0;
	f->seen_data_not_found = 0;
	f->seen_route_not_found = 0;

	f->seen_data_chunk = 0;
	f->seen_length = 0;
	f->seen_data = 0;
	f->seen_data_complete = 0;

	f->seen_success = 0;
	f->seen_pending = 0;
	f->seen_key_collision = 0;
	f->seen_size_error = 0;

	return 0;
}

int fcp_hello(fcp_t *f)
{
	pid_t pid = getpid();
	char *reply = calloc(MAXBUFF, 1);
	int rc = -1;

	if (NULL == reply) {
		LOG(L_ERROR,("{%d} barf: calloc(%d,%d) failure (%s)\n",
			pid, MAXBUFF, 1, strerror(errno)));
		goto bailout;
	}
	if (0 != sk_write(&f->conn, "\0\0\0\2", 4)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent 00 00 00 02\n", pid));
	if (0 != sk_printf(&f->conn, "ClientHello\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent ClientHello\n", pid));
	if (0 != sk_printf(&f->conn, "EndMessage\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent EndMessage\n", pid));

	fcp_reset(f, 1);
	while (0 == (rc = sk_gets(&f->conn, reply, MAXBUFF)) &&
		0 == f->seen_end_message) {
		char *var, *val;
		var = strtok(reply, "=\r\n");
		val = strtok(NULL, "\r\n");
		if (NULL == var) {
			LOG(L_MINOR,("{%d} rcvd '%s'\n", pid, reply));
		} else if (0 == strcasecmp(var, "NodeHello")) {
			f->seen_node_hello = 1;
			LOG(L_DEBUG,("{%d} rcvd NodeHello\n", pid));
		} else if (0 == strcasecmp(var, "Protocol")) {
			LOG(L_DEBUG,("{%d} rcvd Protocol=%d\n", pid, val));
			if (f->seen_protocol)
				free(f->seen_protocol);
			f->seen_protocol = strdup(val);
		} else if (0 == strcasecmp(var, "Node")) {
			LOG(L_DEBUG,("{%d} rcvd Node=%s\n", pid, val));
			if (f->seen_node)
				free(f->seen_node);
			f->seen_node = strdup(val);
		} else if (0 == strcasecmp(var, "EndMessage")) {
			LOG(L_DEBUG,("{%d} rcvd EndMessage\n", pid));
			f->seen_end_message = 1;
		}
	}
	rc = (f->seen_node_hello && f->seen_end_message) ? 0 : -1;

bailout:
	if (NULL != reply) {
		free(reply);
		reply = NULL;
	}
	return rc;
}

int fcp_key(fcp_t *f)
{
	pid_t pid = getpid();
	char *reply = calloc(MAXBUFF, 1);
	int rc = -1;

	if (NULL == reply) {
		LOG(L_ERROR,("{%d} barf: calloc(%d,%d) failure (%s)\n",
			pid, MAXBUFF, 1, strerror(errno)));
		goto bailout;
	}
	if (0 != sk_write(&f->conn, "\0\0\0\2", 4)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent 00 00 00 02\n", pid));
	if (0 != sk_printf(&f->conn, "GenerateSVKPair\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent GenerateSVKPair\n", pid));
	if (0 != sk_printf(&f->conn, "EndMessage\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent EndMessage\n", pid));

	fcp_reset(f, 0);
	while (0 == (rc = sk_gets(&f->conn, reply, MAXBUFF)) &&
		0 == f->seen_end_message) {
		char *var, *val;
		var = strtok(reply, "=\r\n");
		val = strtok(NULL, "\r\n");
		if (NULL == var) {
			LOG(L_MINOR,("{%d} rcvd '%s'\n", pid, reply));
		} else if (0 == strcasecmp(var, "Success")) {
			f->seen_success = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd Success\n", pid));
		} else if (0 == strcasecmp(var, "PublicKey")) {
			if (f->public_key) {
				free(f->public_key);
				f->public_key = NULL;
			}
			f->public_key = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd PublicKey=%s\n",
				pid, f->public_key));
		} else if (0 == strcasecmp(var, "PrivateKey")) {
			if (f->private_key) {
				free(f->private_key);
				f->private_key = NULL;
			}
			f->private_key = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd PrivateKey=%s\n",
				pid, f->private_key));
		} else if (0 == strcasecmp(var, "EndMessage")) {
			LOG(L_DEBUG,("{%d} rcvd EndMessage\n", pid));
			f->seen_end_message = 1;
		}
	}
	rc = (f->seen_success && f->seen_end_message) ? 0 : -1;

bailout:
	if (NULL != reply) {
		free(reply);
		reply = NULL;
	}
	return rc;
}

int fcp_get(fcp_t *f, const char *uri, int htl)
{
	pid_t pid = getpid();
	char *reply = calloc(MAXBUFF, 1);
	int rc = -1;

	if (NULL == reply) {
		LOG(L_ERROR,("barf: calloc(%d,%d) failure (%s)\n",
			MAXBUFF, 1, strerror(errno)));
		goto bailout;
	}
	if (0 != sk_write(&f->conn, "\0\0\0\2", 4)) {
		LOG(L_ERROR,("sending fcp control failed (%s)\n",
			strerror(errno)));
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent 00 00 00 02\n", pid));
	if (0 != sk_printf(&f->conn, "ClientGet\n")) {
		LOG(L_ERROR,("sending ClientGet failed (%s)\n",
			strerror(errno)));
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent ClientGet\n", pid));
	if (htl > 0) {
		if (0 != sk_printf(&f->conn, "HopsToLive=%x\n", htl)) {
			LOG(L_ERROR,("sending HopsToLive=%x failed (%s)\n",
				htl, strerror(errno)));
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent HopsToLive=%x\n", pid, htl));
	}
	if (0 != sk_printf(&f->conn, "URI=%s\n", uri)) {
		LOG(L_ERROR,("sending URI=%s failed (%s)\n",
			uri, strerror(errno)));
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent URI=%s\n", pid, uri));
	if (0 != sk_printf(&f->conn, "EndMessage\n")) {
		LOG(L_ERROR,("sending EndMessage failed (%s)\n",
			strerror(errno)));
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent EndMessage\n", pid));

	do {
		fcp_reset(f, 1);

		while (0 == (rc = sk_gets(&f->conn, reply, MAXBUFF))) {
			char *var, *val;

			var = strtok(reply, "=\r\n");
			val = strtok(NULL, "\r\n");
			if (NULL == var) {
				LOG(L_MINOR,("{%d} rcvd '%s'\n", pid, reply));
			} else if (0 == strcasecmp(var, "DataFound")) {
				f->seen_data_found = 1;
				f->seen_end_message = 0;
				LOG(L_DEBUG,("{%d} rcvd DataFound\n", pid));
			} else if (0 == strcasecmp(var, "DataLength")) {
				f->datasize = strtoul(val, NULL, 16);
				LOG(L_DEBUG,("{%d} rcvd DataLength=%x\n",
					pid, f->datasize));
			} else if (0 == strcasecmp(var, "MetadataLength")) {
				f->metasize = strtoul(val, NULL, 16);
				LOG(L_DEBUG,("{%d} rcvd MetadataLength=%x\n",
					pid, f->metasize));
			} else if (0 == strcasecmp(var, "EndMessage")) {
				f->seen_end_message = 1;
				LOG(L_DEBUG,("{%d} rcvd EndMessage\n", pid));
			} else if (0 == strcasecmp(var, "URIError")) {
				f->seen_uri_error = 1;
				f->seen_end_message = 0;
				LOG(L_DEBUG,("{%d} rcvd URIError\n", pid));
			} else if (0 == strcasecmp(var, "Restarted")) {
				f->seen_restarted = 1;
				f->seen_end_message = 0;
				LOG(L_DEBUG,("{%d} rcvd Restarted\n", pid));
			} else if (0 == strcasecmp(var, "DataNotFound")) {
				f->seen_data_not_found = 1;
				f->seen_end_message = 0;
				LOG(L_DEBUG,("{%d} rcvd DataNotFound\n", pid));
			} else if (0 == strcasecmp(var, "RouteNotFound")) {
				f->seen_route_not_found = 1;
				f->seen_end_message = 0;
				LOG(L_DEBUG,("{%d} rcvd RouteNotFound\n", pid));
			}
			if (0 != f->seen_end_message) {
				if (0 != f->seen_data_found)
					break;
				if (0 != f->seen_uri_error)
					break;
				if (0 != f->seen_data_not_found)
					break;
				if (0 != f->seen_route_not_found)
					break;
				if (0 != f->seen_restarted)
					break;
			}
		}

		if (0 != rc) {
			LOG(L_DEBUG,("bail: rc = %d\n", rc));
			goto bailout;
		}
		if (0 == f->seen_end_message) {
			LOG(L_DEBUG,("bail: did not see EndMessage\n"));
			goto bailout;
		}
		if (0 != f->seen_uri_error) {
			LOG(L_DEBUG,("bail: seen URIError\n"));
			goto bailout;
		}
		if (0 != f->seen_data_not_found) {
			LOG(L_DEBUG,("bail: seen DataNotFound\n"));
			goto bailout;
		}
		if (0 != f->seen_route_not_found) {
			LOG(L_DEBUG,("bail: seen RouteNotFound\n"));
			goto bailout;
		}

		if (f->metasize > 0) {
			f->meta = calloc(f->metasize + 31, 1);
			if (NULL == f->meta) {
				LOG(L_DEBUG,("bail: meta alloc(%d) failed\n",
					f->metasize));
				goto bailout;
			}
		}
		if (f->datasize > 0) {
			if (f->metasize > f->datasize) {
				LOG(L_DEBUG,("bail: metasize > datasize\n"));
				goto bailout;
			}
			f->datasize -= f->metasize;
			if (f->datasize > 0) {
				f->data = calloc(f->datasize + 31, 1);
				if (NULL == f->data) {
					LOG(L_DEBUG,("bail: data alloc(%d) failed\n",
						f->datasize));
					goto bailout;
				}
			}
		}

		f->seen_uri_error = 0;
		f->seen_restarted = 0;
		f->seen_end_message = 0;
		while (0 == f->seen_restarted && 0 == f->seen_data_complete) {
			f->seen_data_chunk = 0;
			f->seen_length = 0;
			f->seen_data = 0;
			while (0 == (rc = sk_gets(&f->conn, reply, MAXBUFF))) {
				char *var, *val;
	
				var = strtok(reply, "=\r\n");
				val = strtok(NULL, "\r\n");
				if (NULL == var) {
					LOG(L_MINOR,("{%d} rcvd '%s'\n",
						pid, reply));
				} else if (0 == strcasecmp(var, "DataChunk")) {
					f->seen_data_chunk = 1;
					LOG(L_DEBUG,("{%d} rcvd DataChunk\n",
						pid));
				} else if (0 == strcasecmp(var, "Length")) {
					f->seen_length = strtoul(val, NULL, 16);
					LOG(L_DEBUG,("{%d} rcvd Length=%x\n",
						pid, f->seen_length));
				} else if (0 == strcasecmp(var, "Data")) {
					f->seen_data = 1;
					LOG(L_DEBUG,("{%d} rcvd Data\n",
						pid));
				} else if (0 == strcasecmp(var, "Restarted")) {
					f->seen_restarted = 1;
					LOG(L_DEBUG,("{%d} rcvd Restarted\n",
						pid));
				}
				if (0 != f->seen_end_message || 0 != f->seen_data) {
					if (0 != f->seen_data_chunk)
						break;
					if (0 != f->seen_restarted)
						break;
				}
			}
			if (0 != f->seen_restarted) {
				LOG(L_DEBUG,("bail: seen Restarted\n"));
				continue;
			}
			if (0 == f->seen_data_chunk) {
				LOG(L_DEBUG,("bail: did not see DataChunk\n"));
				goto bailout;
			}
			if (0 == f->seen_length) {
				LOG(L_DEBUG,("bail: did not see Length=\n"));
				goto bailout;
			}
			if (0 == f->seen_data) {
				LOG(L_DEBUG,("bail: did not see Data\n"));
				goto bailout;
			}
			if (f->metaoffs < f->metasize) {
				size_t size = f->seen_length;
				if (size > f->metasize - f->metaoffs) {
					size = f->metasize - f->metaoffs;
				}
				LOG(L_DEBUG,("reading max 0x%x bytes metadata\n",
					size));
				if (0 != sk_read(&f->conn, &f->meta[f->metaoffs], size)) {
					LOG(L_DEBUG,("bail: read(%d) failed\n", size));
					goto bailout;
				}
				f->metaoffs += size;
				f->seen_length -= size;
				LOG(L_DEBUG,("{%d} rcvd [%x bytes metadata]\n",
					pid, size));
				if (0 == f->datasize &&
					f->metaoffs >= f->metasize) {
					LOG(L_DEBUG,("{%d} data complete\n",
						pid));
					f->seen_data_complete = 1;
				}
			}
			if (f->dataoffs < f->datasize) {
				size_t size = f->seen_length;
				if (size > f->datasize - f->dataoffs) {
					size = f->datasize - f->dataoffs;
				}
				LOG(L_DEBUG,("reading max 0x%x bytes data\n",
					size));
				if (0 != sk_read(&f->conn, &f->data[f->dataoffs], size)) {
					LOG(L_DEBUG,("bail: read(%d) failed\n", size));
					goto bailout;
				}
				f->dataoffs += size;
				f->seen_length -= size;
				LOG(L_DEBUG,("{%d} rcvd [%x bytes data]\n",
					pid, size));
				if (f->dataoffs >= f->datasize) {
					LOG(L_DEBUG,("{%d} data complete\n",
						pid));
					f->seen_data_complete = 1;
				}
			}
		}
	} while (0 != f->seen_restarted);

bailout:
	if (NULL != reply) {
		free(reply);
		reply = NULL;
	}
	LOG(L_DEBUG,("{%d} return %d\n",
		pid, rc));
	return rc;
}

void fcp_signal(int sig)
{
	signal(sig, SIG_DFL);
	if (SIGHUP == sig) {
		exit(1);
	}
}

void set_signal(int sig, void (*my_signal_handler)(int))
{
	struct sigaction sa;
	signal(sig, my_signal_handler);
	sigaction(sig, (struct sigaction *)0, &sa);
#ifdef	SA_RESTART
	sa.sa_flags |= SA_RESTART;
#endif
#ifdef	SA_INTERRUPT
	sa.sa_flags &= ~SA_INTERRUPT;
#endif
	sigaction(sig, &sa, (struct sigaction *)0);
}

int fcp_put(fcp_t *f, const char *uri, int htl, int delete_local)
{
	pid_t pid = getpid();
	char *reply = calloc(MAXBUFF, 1);
	size_t size;
	int rc = -1;

	set_signal(SIGPIPE, fcp_signal);

	if (NULL == reply) {
		LOG(L_ERROR,("{%d} barf: calloc(%d,%d) failure (%s)\n",
			pid, MAXBUFF, 1, strerror(errno)));
		goto bailout;
	}
	if (0 != sk_write(&f->conn, "\0\0\0\2", 4)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent 00 00 00 02\n", pid));

	if (0 != sk_printf(&f->conn, "ClientPut\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent ClientPut\n", pid));

	if (htl > 0) {
		if (0 != sk_printf(&f->conn, "HopsToLive=%x\n", htl)) {
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent HopsToLive=%x\n", pid, htl));
	}

	if (delete_local) {
		if (0 != sk_printf(&f->conn, "RemoveLocalKey=true\n")) {
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent RemoveLocalKey=true\n", pid));
	}

	if (0 != sk_printf(&f->conn, "URI=%s\n", uri)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent URI=%s\n", pid, uri));

	size = f->datasize + f->metasize;
	if (0 != sk_printf(&f->conn, "DataLength=%x\n", size)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent DataLength=%x\n", pid, size));

	if (f->metasize > 0) {
		size = f->metasize;
		if (0 != sk_printf(&f->conn, "MetadataLength=%x\n", size)) {
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent MetadataLength=%x\n", pid, size));
	}
	if (0 != sk_printf(&f->conn, "Data\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent Data\n", pid));

	if (f->metasize > 0) {
		if (0 != (rc = sk_write(&f->conn, f->meta, f->metasize))) {
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent [0x%x bytes metadata]\n",
			pid, f->metasize));
	}

	if (f->datasize > 0) {
		if (0 != (rc = sk_write(&f->conn, f->data, f->datasize))) {
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent [0x%x bytes data]\n",
			pid, f->datasize));
	}

	f->seen_node_hello = 0;
	f->seen_success = 0;
	f->seen_uri_error = 0;
	f->seen_restarted = 0;
	f->seen_pending = 0;
	f->seen_route_not_found = 0;
	f->seen_size_error = 0;
	if (NULL != f->public_key) {
		free(f->public_key);
		f->public_key = NULL;
	}
	if (NULL != f->private_key) {
		free(f->private_key);
		f->private_key = NULL;
	}
	f->seen_end_message = 0;
	while (0 == (rc = sk_gets(&f->conn, reply, MAXBUFF))) {
		char *var, *val;

		var = strtok(reply, "=\r\n");
		val = strtok(NULL, "\r\n");
		if (NULL == var) {
			LOG(L_MINOR,("{%d} rcvd '%s'\n", pid, reply));
		} else if (0 == strcasecmp(var, "Success")) {
			f->seen_success = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd Success\n", pid));
		} else if (0 == strcasecmp(var, "URI")) {
			if (f->uri) {
				free(f->uri);
				f->uri = NULL;
			}
			f->uri = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd URI=%s\n", pid, f->uri));
		} else if (0 == strcasecmp(var, "PublicKey")) {
			if (f->public_key) {
				free(f->public_key);
				f->public_key = NULL;
			}
			f->public_key = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd PublicKey=%s\n",
				pid, f->public_key));
		} else if (0 == strcasecmp(var, "PrivateKey")) {
			if (f->private_key) {
				free(f->private_key);
				f->private_key = NULL;
			}
			f->private_key = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd PrivateKey=%s\n",
				pid, f->private_key));
		} else if (0 == strcasecmp(var, "EndMessage")) {
			f->seen_end_message = 1;
			LOG(L_DEBUG,("{%d} rcvd EndMessage\n", pid));
		} else if (0 == strcasecmp(var, "URIError")) {
			f->seen_uri_error = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd URIError\n", pid));
		} else if (0 == strcasecmp(var, "Restarted")) {
			f->seen_restarted = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd Restarted\n", pid));
		} else if (0 == strcasecmp(var, "RouteNotFound")) {
			f->seen_route_not_found = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd RouteNotFound\n", pid));
		} else if (0 == strcasecmp(var, "Pending")) {
			f->seen_pending = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd Pending\n", pid));
		} else if (0 == strcasecmp(var, "KeyCollision")) {
			f->seen_key_collision = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd KeyCollision\n", pid));
		} else if (0 == strcasecmp(var, "SizeError")) {
			f->seen_size_error = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd SizeError\n", pid));
		}
		if (0 != f->seen_end_message) {
			if (0 != f->seen_success)
				break;
			if (0 != f->seen_uri_error)
				break;
			if (0 != f->seen_route_not_found)
				break;
			if (0 != f->seen_key_collision)
				break;
			if (0 != f->seen_size_error)
				break;
		}
	}

	if (0 != rc) {
		LOG(L_DEBUG,("bail: rc = %d\n", rc));
		goto bailout;
	}
	rc = -1;
	if (0 == f->seen_end_message) {
		LOG(L_DEBUG,("bail: did not see EndMessage\n"));
		goto bailout;
	}
	if (0 != f->seen_uri_error) {
		LOG(L_DEBUG,("bail: seen URIError\n"));
		goto bailout;
	}
	if (0 != f->seen_data_not_found) {
		LOG(L_DEBUG,("bail: seen DataNotFound\n"));
		goto bailout;
	}
	if (0 != f->seen_route_not_found) {
		LOG(L_DEBUG,("bail: seen RouteNotFound\n"));
		goto bailout;
	}
	rc = 0;

bailout:
	if (NULL != reply) {
		free(reply);
		reply = NULL;
	}
	return rc;
}

int fcp_chk(fcp_t *f)
{
	pid_t pid = getpid();
	char *reply = calloc(MAXBUFF, 1);
	size_t size;
	int rc = -1;

	if (NULL == reply) {
		LOG(L_ERROR,("{%d} barf: calloc(%d,%d) failure (%s)\n",
			pid, MAXBUFF, 1, strerror(errno)));
		goto bailout;
	}
	if (0 != sk_write(&f->conn, "\0\0\0\2", 4)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent 00 00 00 02\n", pid));

	if (0 != sk_printf(&f->conn, "GenerateCHK\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent GenerateCHK\n", pid));

	size = f->datasize + f->metasize;
	if (0 != sk_printf(&f->conn, "DataLength=%x\n", size)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent DataLength=%x\n", pid, size));

	if (f->metasize > 0) {
		size = f->metasize;
		if (0 != sk_printf(&f->conn, "MetadataLength=%x\n", size)) {
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent MetadataLength=%x\n",
			pid, size));
	}
	if (0 != sk_printf(&f->conn, "Data\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent Data\n", pid));

	if (f->metasize > 0) {
		if (0 != (rc = sk_write(&f->conn, f->meta, f->metasize))) {
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent [0x%x bytes metadata]\n",
			pid, f->metasize));
	}

	if (f->datasize > 0) {
		if (0 != (rc = sk_write(&f->conn, f->data, f->datasize))) {
			goto bailout;
		}
		LOG(L_DEBUG,("{%d} sent [0x%x bytes data]\n",
			pid, f->datasize));
	}

	f->seen_success = 0;
	f->seen_restarted = 0;
	f->seen_pending = 0;
	if (NULL != f->public_key) {
		free(f->public_key);
		f->public_key = NULL;
	}
	if (NULL != f->private_key) {
		free(f->private_key);
		f->private_key = NULL;
	}
	f->seen_end_message = 0;
	while (0 == (rc = sk_gets(&f->conn, reply, MAXBUFF))) {
		char *var, *val;

		var = strtok(reply, "=\r\n");
		val = strtok(NULL, "\r\n");
		if (NULL == var) {
			LOG(L_DEBUG,("{%d} rcvd '%s'\n", pid, reply));
		} else if (0 == strcasecmp(var, "Success")) {
			f->seen_success = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd Success\n", pid));
		} else if (0 == strcasecmp(var, "URI")) {
			if (f->uri) {
				free(f->uri);
				f->uri = NULL;
			}
			f->uri = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd URI=%s\n", pid, f->uri));
		} else if (0 == strcasecmp(var, "PublicKey")) {
			if (f->public_key) {
				free(f->public_key);
				f->public_key = NULL;
			}
			f->public_key = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd PublicKey=%s\n",
				pid, f->public_key));
		} else if (0 == strcasecmp(var, "PrivateKey")) {
			if (f->private_key) {
				free(f->private_key);
				f->private_key = NULL;
			}
			f->private_key = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd PrivateKey=%s\n",
				pid, f->private_key));
		} else if (0 == strcasecmp(var, "EndMessage")) {
			f->seen_end_message = 1;
			LOG(L_DEBUG,("{%d} rcvd EndMessage\n", pid));
		} else if (0 == strcasecmp(var, "Pending")) {
			f->seen_pending = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd Pending\n", pid));
		}
		if (0 != f->seen_end_message) {
			if (0 != f->seen_success)
				break;
		}
	}

	if (0 != rc) {
		LOG(L_DEBUG,("bail: rc = %d\n", rc));
		goto bailout;
	}
	rc = -1;
	if (0 == f->seen_end_message) {
		LOG(L_DEBUG,("bail: did not see EndMessage\n"));
		goto bailout;
	}
	rc = 0;

bailout:
	if (NULL != reply)
		free(reply);
	return rc;
}

int fcp_del(fcp_t *f, const char *uri)
{
	pid_t pid = getpid();
	char *reply = calloc(MAXBUFF, 1);
	int rc = -1;

	if (NULL == reply) {
		LOG(L_ERROR,("{%d} barf: calloc(%d,%d) failure (%s)\n",
			pid, MAXBUFF, 1, strerror(errno)));
		goto bailout;
	}
	if (0 != sk_write(&f->conn, "\0\0\0\2", 4)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent 00 00 00 02\n", pid));

	if (0 != sk_printf(&f->conn, "ClientDelete\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent ClientDelete\n", pid));

	if (0 != sk_printf(&f->conn, "URI=%s\n", uri)) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent URI=%s\n", pid, uri));

	if (0 != sk_printf(&f->conn, "EndMessage\n")) {
		goto bailout;
	}
	LOG(L_DEBUG,("{%d} sent EndMessage\n", pid));

	f->seen_success = 0;
	f->seen_restarted = 0;
	f->seen_pending = 0;
	if (NULL != f->public_key) {
		free(f->public_key);
		f->public_key = NULL;
	}
	if (NULL != f->private_key) {
		free(f->private_key);
		f->private_key = NULL;
	}
	f->seen_end_message = 0;
	while (0 == (rc = sk_gets(&f->conn, reply, MAXBUFF))) {
		char *var, *val;

		var = strtok(reply, "=\r\n");
		val = strtok(NULL, "\r\n");
		if (NULL == var) {
			LOG(L_DEBUG,("{%d} rcvd '%s'\n",
				pid, reply));
		} else if (0 == strcasecmp(var, "Success")) {
			f->seen_success = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd Success\n", pid));
		} else if (0 == strcasecmp(var, "URI")) {
			if (f->uri) {
				free(f->uri);
				f->uri = NULL;
			}
			f->uri = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd URI=%s\n", pid, f->uri));
		} else if (0 == strcasecmp(var, "PublicKey")) {
			if (f->public_key) {
				free(f->public_key);
				f->public_key = NULL;
			}
			f->public_key = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd PublicKey=%s\n",
				pid, f->public_key));
		} else if (0 == strcasecmp(var, "PrivateKey")) {
			if (f->private_key) {
				free(f->private_key);
				f->private_key = NULL;
			}
			f->private_key = strdup(val);
			LOG(L_DEBUG,("{%d} rcvd PrivateKey=%s\n",
				pid, f->private_key));
		} else if (0 == strcasecmp(var, "EndMessage")) {
			f->seen_end_message = 1;
			LOG(L_DEBUG,("{%d} rcvd EndMessage\n", pid));
		} else if (0 == strcasecmp(var, "Pending")) {
			f->seen_pending = 1;
			f->seen_end_message = 0;
			LOG(L_DEBUG,("{%d} rcvd Pending\n", pid));
		}
		if (0 != f->seen_end_message) {
			if (0 != f->seen_success)
				break;
		}
	}

	if (0 != rc) {
		LOG(L_DEBUG,("bail: rc = %d\n", rc));
		goto bailout;
	}
	rc = -1;
	if (0 == f->seen_end_message) {
		LOG(L_DEBUG,("bail: did not see EndMessage\n"));
		goto bailout;
	}
	rc = 0;

bailout:
	if (NULL != reply) {
		free(reply);
		reply = NULL;
	}
	return rc;
}
