/*****************************************************************************
 *  FTools - Freenet Client Protocol Tools
 *
 *  Copyright (C) 2002 Juergen Buchmueller <juergen@pullmoll.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *****************************************************************************/
#include "ftsock.h"
#include "ftlog.h"

#ifdef hpux
# define socklen_t int
#endif

int fork2(void)
{
	pid_t pid;
	int status;

	if (0 == (pid = fork())) {
		switch (fork()) {
		case -1:
			_exit(errno);
		case 0:
			return 0;
		default:
			_exit(0);
		}
	}

	if (pid < 0 || waitpid(pid, &status, 0) < 0) {
		return -1;
	}

	if (WIFEXITED(status)) {
		if (0 == WEXITSTATUS(status)) {
			return 1;
		} else {
			errno = WEXITSTATUS(status);
		}
	} else {
		errno = EINTR;
	}
	return -1;
}

int sk_outgoing(conn_t *conn, const char *hostname, int port)
{
	struct in_addr addr;
	struct hostent *host;
	int sk;

	if (inet_aton(hostname, &addr) != 0) {
		host = gethostbyname(hostname);
		if (NULL == host) {
			return -1;
		}
		addr.s_addr = ((struct in_addr *) *host->h_addr_list)->s_addr;
	}

	memset((char *)&conn->addr, 0, sizeof(conn->addr));
	conn->sk = -1;
	conn->addr.sin_family = AF_INET;
	conn->addr.sin_port = htons(port);
	conn->addr.sin_addr.s_addr = addr.s_addr;

	sk = socket(AF_INET, SOCK_STREAM, 0);
	if (sk < 0) {
		return -1;
	}

	if (connect(sk, (struct sockaddr *)&conn->addr, sizeof(conn->addr)) < 0) {
		return -1;
	}

	if (fcntl(sk, F_SETFL, O_NONBLOCK) < 0) {
		return -1;
	}

	conn->sk = sk;
	return 0;
}

int sk_incoming(const char *hostname, int port, int (*child)(conn_t *conn))
{
	struct in_addr addr;
	struct hostent *host;
	struct sockaddr_in in;
	int sk, arg = 1;

	if (inet_aton(hostname, &addr) != 0) {
		host = gethostbyname(hostname);
		if (NULL == host) {
			return -1;
		}
		addr.s_addr = ((struct in_addr *) *host->h_addr_list)->s_addr;
	}

	memset((char *)&in, 0, sizeof(in));
	in.sin_family = AF_INET;
	in.sin_port = htons(port);
	in.sin_addr.s_addr = addr.s_addr;

	sk = socket(AF_INET, SOCK_STREAM, 0);
	if (sk < 0) {
		return -1;
	}

	setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, (char *)&arg, sizeof(arg));

	if (bind(sk, (struct sockaddr *)&in, sizeof(in)) < 0) {
		close(sk);
		return -1;
	}

	for (;;) {
		int sk2, rc;

		if (0 != listen(sk, 5)) {
			close(sk);
			return -1;
		}

		sk2 = -1;
		while (sk2 < 0) {
			conn_t *conn = NULL;
			socklen_t socklen = sizeof(conn->addr);
			int stat;

			conn = calloc(sizeof(conn_t), 1);
			if (NULL == conn) {
				close(sk);
				return -1;
			}

			sk2 = accept(sk, (struct sockaddr *)&conn->addr, &socklen);

			switch (fork2()) {
			case -1:
				return -1;
			case 0:
				close(sk);
				conn->sk = sk2;
				rc = (*child)(conn);
				close(sk2);
				sk_shut(conn);
				free(conn);
				_exit(rc);
			default:
				sk_shut(conn);
				free(conn);
				while (wait(&stat) >= 0)
					;
				close(sk2);
			}
		}
	}
	close(sk);
	return 0;
}

int sk_write(conn_t *conn, const void *buffer, size_t size)
{
	const char *buff = (const char *)buffer;
	size_t offs = 0;

	while (offs < size) {
		ssize_t count;
		size_t limit = size - offs;
		struct timeval tv = {60,0};	/* 60 seconds timeout */
		fd_set wrfds, exfds;
		int sk;

		sk = conn->sk;
		if (-1 == conn->sk) {
			return -1;
		}

		FD_ZERO(&wrfds);
		FD_SET(sk, &wrfds);
		FD_ZERO(&exfds);
		FD_SET(sk, &exfds);
		if (-1 == select(sk+1, NULL, &wrfds, &exfds, &tv)) {
			return -1;
		}
		if (FD_ISSET(sk, &exfds)) {
			sk_shut(conn);
			return -1;
		}

		count = send(sk, &buff[offs], limit, 0);
		if (-1 == count && errno == EAGAIN) {
			usleep(25000);
			continue;
		} else if (count < 1) {
			sk_shut(conn);
			return -1;
		}
		offs += count;
	}

	return 0;
}

int sk_read(conn_t *conn, void *buffer, size_t size)
{
	char *buff = (char *)buffer;
	size_t offs = 0;
	int rc = 0;

	while (offs < size && rc >= 0) {
		ssize_t count;
		size_t limit = size - offs;
		struct timeval tv = {600,0};	/* 600 seconds timeout */
		fd_set rdfds, exfds;
		int sk;

		sk = conn->sk;
		if (-1 == conn->sk) {
			return -1;
		}

		FD_ZERO(&rdfds);
		FD_SET(sk, &rdfds);
		FD_ZERO(&exfds);
		FD_SET(sk, &exfds);
		if (-1 == select(sk+1, &rdfds, NULL, &exfds, &tv)) {
			rc = -1;
		}
		if (FD_ISSET(sk, &exfds)) {
			sk_shut(conn);
			rc = -1;
		}
		if (rc >= 0 && FD_ISSET(sk, &rdfds)) {
			count = recv(sk, &buff[offs], limit, 0);
			if (-1 == count && errno == EAGAIN) {
				usleep(25000);
				continue;
			} else if (count < 1) {
				sk_shut(conn);
				return -1;
			} else if (0 == offs && 0 == count) {
				sk_shut(conn);
				errno = EBADF;
				return -1;
			}
			offs += count;
		}
	}

	if (rc >= 0) {
		rc = 0;
	}
	return rc;
}

int sk_gets(conn_t *conn, char *buff, size_t size)
{
	size_t offs = 0;
	int rc = 0;

	while (size > 0 && rc >= 0) {
		ssize_t count;
		struct timeval tv = {600,0};	/* 600 seconds timeout */
		fd_set rdfds, exfds;
		int sk;

		sk = conn->sk;
		if (-1 == conn->sk) {
			return -1;
		}

		FD_ZERO(&rdfds);
		FD_SET(sk, &rdfds);
		FD_ZERO(&exfds);
		FD_SET(sk, &exfds);
		if (-1 == select(sk+1, &rdfds, NULL, &exfds, &tv)) {
			LOG(L_ERROR,("select() failed (%s)\n", strerror(errno)));
			rc = -1;
		}
		if (FD_ISSET(sk, &exfds)) {
			LOG(L_ERROR,("sk is in exfds (%s)\n", strerror(errno)));
			rc = -1;
		}

		if (rc >= 0 && FD_ISSET(sk, &rdfds)) {
			count = recv(sk, &buff[offs], 1, 0);
			if (-1 == count && errno == EAGAIN) {
				usleep(25000);
				continue;
			}
			if (0 == offs && 0 == count && errno != EAGAIN) {
				LOG(L_ERROR,("EOF: %s\n", strerror(errno)));
				rc = -1;
			} else if (count < 1) {
				LOG(L_ERROR,("recv() failed (%s)\n", strerror(errno)));
				rc = -1;
			}
			size -= count;
			if (buff[offs++] == '\n')
				break;
			size--;
		}
	}

	buff[offs] = '\0';
	return rc;
}

int sk_printf(conn_t *conn, const char *fmt, ...)
{
	char *buff = calloc(32768,1);
	va_list ap;
	size_t len;

	if (NULL == buff) {
		return -1;
	}

	va_start(ap, fmt);
	len = vsnprintf(buff, 32768, fmt, ap);
	va_end(ap);

	sk_write(conn, buff, len);

	free(buff);
	return 0;
}

int sk_shut(conn_t *conn)
{
	int rc = -1;

	if (NULL == conn) {
		errno = EINVAL;
		return rc;
	}

	if (-1 == conn->sk) {
		errno = EBADF;
		return rc;
	}

	if (0 != (rc = shutdown(conn->sk, 2))) {
		return rc;
	}

	if (0 != (rc = close(conn->sk))) {
		return rc;
	}

	conn->sk = -1;
	return rc;
}
